unit HandleReduction;

{$mode objfpc}{$H+}

interface

uses SysUtils, Basics, BraidBasics;

{-----------------------------------------------}
{               HANDLE REDUCTION                }
{-----------------------------------------------}

function BEquiv(B, BB: braid): boolean;
procedure BStepByStepHReduce (var B: Braid);
procedure BFullHReduce (var B: Braid);
procedure BFullHFReduce (var B: Braid);
var BNbRedSteps: longInt;
procedure BQuickFullHFReduce (var B: Braid; threshold: integer);
var BReduced: boolean;

{-----------------------------------------------}
{ BRAID ORDERING }
{-----------------------------------------------}

{ BOccurrences(B) is a string saying with which sign(s) sigma_i occurs in B, in view of orderings. }
function BOccurrences (B: braid): string;
// "D" for the original D-ordering, "F" for the flipped version, "U" for Dubrovins
var BOrderOption: char;
function BOrderName: string;
{ BReduce(B) reduces the handles of B in function of BOrderOption. }
procedure BReduce (var B: braid);
function BSign (B: braid): char;
{ BVisibleSign(B) is the sign of B w.r.t. BOrderOption when directly visible. }
function BVisibleSign (B: braid): char;

implementation

{-----------------------------------------------}
{               HANDLE REDUCTION                }
{-----------------------------------------------}

{ BEquiv(B, BB) is true iff B and BB represent the same braid. }
function BEquiv(B, BB: braid): boolean;
begin
	BInv(B);
	BConcat(B, BB);
	BFullHFReduce(B);
	BEquiv:= (B.L = 0);
end; { BEquiv }

{ Handle reduction consists in deleting the alternations X..x or }
{ x..X in order to produce BReduced words, ie, words where the }
{ generator with least index occurs only positively, or only }
{ negatively. }

{ The plan is as follows: }
{ BASIC PROCEDURES describe how a handle is choosen, and how it is }
{ BReduced; }
{ MAIN PROCEDURES then concatenates the handle reductions according }
{ to the various possible modes. }

{ BASIC PROCEDURES: 1. CHOICE OF A HANDLE }

{ findHandle(B, p, q, i) lets p..q be the leftmost handle of B, }
{ if any, and 0..x otherwise; i is the index of the handle that }
{ has been found. }
{ findNestedHandle(B, p, q, i) is similar for a nested handle. }

procedure findHandle (B: Braid; var p, q, i: Integer);
var
	Last: array[1..BMaxWidth] of Integer;
	j, k, e: Integer;
begin
	for k := 1 to B.W - 1 do
		Last[k] := 0;
	p := 0;
	q := 1;
	while (p = 0) and (q <= B.L) do
		begin
			i := Abs(B.V[q]);
			if B.V[q] > 0 then
				e := 1
			else
				e := -1;
			if Last[i] * e < 0 then
				p := Abs(Last[i])
			else
				begin
					if B.V[q] > 0 then
						Last[i] := q
					else
						Last[i] := -q;
					for j:= i + 1 to B.W do
          	Last[j] := 0;
					q := q + 1;
				end;
		end;
end; { findHandle }

{ findNestedHandle(B, p, q, i) is similar to findHandle(B, p, q, i), }
{ excepted that one looks for a nested handle. To this end one }
{ appeals to localFindHandle(B, x, y, p, q, i), which lets }
{ p..q be the leftmost i-handle that is nested in x..y. }

procedure localFindHandle (B: Braid; x, y: Integer; var p, q, i: Integer);
var
	r: Integer;
begin
	p := 0;
	q := 0;
	i := B.W;
	r := x;
	repeat
		if Abs(B.V[r]) < Abs(i) then
			begin
				i := B.V[r];
				p := r;
				q := 0;
			end
		else if q = 0 then
			begin
				if B.V[r] = i then
					p := r;
				if B.V[r] = -i then
					q := r;
			end;
		r := r + 1;
	until r > y;
	if i < 0 then
		i := -i;
end; { localFindHandle }

procedure findNestedHandle (B: Braid; var p, q, i: Integer);
var
	pp, qq, ii: Integer;
begin
	q := 0;
	localFindHandle(B, 1, B.L, p, q, i);
	if q = 0 then
		begin
			p := 0;
			exit;
		end;
	ii := i - 1;
	while (q > 0) and (i - 1 = ii) do
		begin
			pp := p;
			qq := q;
			ii := i;
			localFindHandle(B, pp + 1, qq - 1, p, q, i);
		end;
	p := pp;
	q := qq;
	i := ii;
end; { findNestedHandle }

{ BASIC PROCEDURES: 2. REDUCTION OF A HANDLE }

{ The possibilities are GlobalreduceHandle(B, p, q, i), }
{ LocalreduceHandle(B, p, q, i) and ConvexreduceHandle(B, p, q, i). }
{ In each case the i-handle that is supposed to lie in B between }
{ p and q is BReduced according to the selected mode. (the result }
{ is correct only if such a handle exists. }

function BShift(i, e: integer): integer;
begin
	if i > 0 then
		BShift:= i + e
	else
		BShift:= i - e;
end; { BShift }

procedure reduceHandle (var B: braid; p, q, i: integer);
var
	m, c, r, mm: Integer;
begin

{ PHASE 1 (preparation): One lets m be the number of blocks of i+1 in }
{ the handle. This m determines the final change of length. }

	m := 0;
	for r := p + 1 to q - 1 do
		if (Abs(B.V[r]) = i + 1) and (Abs(B.V[r + 1]) <> i + 1) then
			m := m + 1;

{ PHASE 2 (actual reduction): One replaces the m blocks of (i+1)^k }
{ with  -c i^k c, where c is (i+1)^±1. The cases m=0 and m=1 are special }
{ because the length does not increase. }

	if m = 0 then
		begin
			B.L := B.L - 2;
			for r := p to q - 2 do
				B.V[r] := B.V[r + 1];
			for r := q - 1 to B.L do
				B.V[r] := B.V[r + 2];
		end
	else
		begin

{ One begins with the direct copy of the letters up to the first }
{ block of 'i+1's.  }

			c := BShift(B.V[p], 1);
			r := p;
			while (Abs(B.V[r + 1]) < i - 1) or (Abs(B.V[r + 1]) > i + 1) do
				begin
					B.V[r] := B.V[r + 1];
					r := r + 1;
				end;
			B.V[r] := -c;
			r := r + 1;
			while Abs(B.V[r]) = i + 1 do
				begin
					B.V[r] := BShift(B.V[r], -1);
					r := r + 1;
				end;

{ Now one makes the Reverseerse copy, ie, the copy of those letters }
{ that are translated to the right. For m=1 only the letters before }
{ q are concerned, for m>1 one has also to translate the end of the }
{ word by 2(m-1) units. }

			mm := 2 * (m - 1);
			if mm > 0 then
				begin
					B.L := B.L + mm;
					for r := B.L downto q + mm + 1 do
						B.V[r] := B.V[r - mm];
				end;
			r := q + mm;
			while mm > 0 do
				begin
					while (Abs(B.V[r - mm - 1]) < i - 1) or (Abs(B.V[r - mm - 1]) > i + 1) do
						begin
							B.V[r] := B.V[r - mm - 1];
							r := r - 1;
						end;
					B.V[r] := c;
					r := r - 1;
					while Abs(B.V[r - mm]) = i + 1 do
						begin
							B.V[r] := BShift(B.V[r - mm], -1);
							r := r - 1;
						end;
					B.V[r] := -c;
					r := r - 1;
					mm := mm - 2;
				end;
			while (Abs(B.V[r - mm - 1]) < i - 1) or (Abs(B.V[r - mm - 1]) > i + 1) do
				begin
					B.V[r] := B.V[r - 1];
					r := r - 1;
				end;
			B.V[r] := c;
		end;
end;	{ reduceHandle }

{ FULL REDUCTION }

{ BFullHReduce is the basic handle reduction. BFullHFReduce includes
in addition a systematic free reduction after each step. }
{ The procedures with "Stat" include counters and the possibility }
{ of displaying the results. }
procedure BFullHReduce (var B: Braid);
var
	p, q, i: Integer;
begin
	BReduced := FALSE;
	BNbRedSteps := 0;
	repeat
		findHandle(B, p, q, i);
		if p = 0 then
			BReduced := TRUE
		else
			reduceHandle(B, p, q, i);
		BNbRedSteps:= BNbRedSteps + 1;
	until BReduced;
end; { BFullHReduce }

procedure BFullHFReduce (var B: Braid);
var p, q, i: integer;
begin
	BReduced := FALSE;
	BNbRedSteps := 0;
	repeat
		FindHandle(B, p, q, i);
		if p = 0 then
			BReduced := True
		else
			reduceHandle(B, p, q, i);
		BFreeReduce(B);
		BNbRedSteps:= BNbRedSteps + 1;
	until BReduced;
end;	// BFullHFReduce

procedure BStepByStepHReduce (var B: Braid);
var p, q, i: integer;
begin
	BReduced := FALSE;
	BNbRedSteps := 0;
	repeat
    write(BNbRedSteps:4, ': ');
		FindHandle(B, p, q, i);
		if p = 0 then
			begin
      	BReduced := TRUE;
        BWriteLn(B);
      end
    else
			begin
        BSpecialWriteLn(B, p, q);
        reduceHandle(B, p, q, i);
      end;
    BNbRedSteps:= BNbRedSteps + 1;
	until BReduced;
end;	// BStepByStepHFReduce

procedure BQuickFullHFReduce (var B: braid; threshold: integer);
var
	BB: braid;
	N, NN: longInt;
begin
	BFreeReduce(B);
	if B.L = 0 then
		exit;
	if B.L <= threshold then
		BFullHFReduce(B)
	else
		begin
			BSplit(B, BB);
			BQuickFullHFReduce(B, threshold);
			N := BNbRedSteps;
			BQuickFullHFReduce(BB, threshold);
			NN := BNbRedSteps;
			BConcat(B, BB);
			BFullHFReduce(B);
			BNbRedSteps := BNbRedSteps + N + NN;
		end;
end; { BQuickFullHFReduce }

{-----------------------------------------------}
{ BRAID ORDERING }
{-----------------------------------------------}

{ BOccurrences(B) is an array saying with which sign(s) sigma_i occurs in B, in view of orderings. }
function BOccurrences (B: braid): string;
var
	w: string[BMaxWidth];
	i, p: integer;
begin
	w:= '';
	for i:= 1 to B.W - 1 do
		w := concat(w, '0');
	for p:= 1 to B.L do
		begin
			if B.V[p] > 0 then
				if w[B.V[p]] = '0' then
					w[B.V[p]] := '+'
				else if w[B.V[p]] = '-' then
					w[B.V[p]] := '?';
			if B.V[p] < 0 then
				if w[-B.V[p]] = '0' then
					w[-B.V[p]] := '-'
				else if w[-B.V[p]] = '+' then
					w[-B.V[p]] := '?';
		end;
	BOccurrences := w;
end; { BOccurrences }

// "D" for the original D-ordering, "F" for the flipped version, "U" for Dubrovins
function BOrderName: string;
begin
	case BOrderOption of
		'D': BOrderName := 'the original D-ordering';
		'F': BOrderName := 'the flipped D-ordering';
		'U': BOrderName := 'the DD-ordering';
		end;
end; { BOrderName }

{ BReduce(B) reduces the handles of B in function of BOrderOption. }
procedure BReduce (var B: braid);
begin
	if BOrderOption <> 'F' then
		BFullHFReduce(B)
	else
		begin
			BFlip(B);
			BFullHFReduce(B);
			BFlip(B);
		end;
end; { BReduce }

function BSign(B: braid): char;
var
	c: char;
begin
	c := BVisibleSign(B);
	if c <> '?' then
		BSign := c
	else
		begin
			BReduce(B);
			BSign := BVisibleSign(B);
		end;
end; { BSign }

function BVisibleSign(B: braid): char;
var
	w: string;
	p: integer;
begin
	w := BOccurrences(B);
	case BOrderOption of
		'D':
			begin
				p := 0;
				repeat p := p+1 until (w[p] <> '0') or (p > length(w));
				if p > length(w) then BVisibleSign := '1'
				else BVisibleSign := w[p];
			end;
		'F':
			begin
				p := length(w)+1;
				repeat p := p-1 until (w[p] <> '0') or (p = 0);
				if p = 0 then BVisibleSign := '1'
				else BVisibleSign := w[p];
			end;
		'U':
			begin
				p := 0;
				repeat p := p+1 until (w[p] <> '0') or (p > length(w));
				if p > length(w) then BVisibleSign := '1'
				else if odd(p) then BVisibleSign := w[p]
				else case w[p] of
					'+': BVisibleSign := '-';
					'-': BVisibleSign := '+';
					'?': BVisibleSign := '?';
					end;
			end;
		end;
end; { BVisibleSign }



end.

